// Copyright (c) Microsoft. All rights reserved.
// Licensed under the MIT license. See LICENSE file in the project root for full license information.
'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
exports.X509Registration = void 0;
const polling_state_machine_1 = require("./polling_state_machine");
const dbg = require("debug");
const azure_iot_common_1 = require("azure-iot-common");
const debugErrors = dbg('azure-iot-provisioning-device:X509Registration:Errors');
/**
 * Client used to run the registration of a device using X509 authentication.
 */
class X509Registration {
    constructor(provisioningHost, idScope, transport, securityClient) {
        this._provisioningHost = provisioningHost;
        this._idScope = idScope;
        this._transport = transport;
        this._securityClient = securityClient;
        this._pollingStateMachine = new polling_state_machine_1.PollingStateMachine(this._transport);
    }
    /**
     * Sets the custom payload for registration that will be sent to the custom allocation policy implemented in an Azure Function.
     *
     * @param payload The payload sent to the provisioning service at registration.
     */
    setProvisioningPayload(payload) {
        this._provisioningPayload = payload;
    }
    register(callback) {
        return (0, azure_iot_common_1.callbackToPromise)((_callback) => {
            /* Codes_SRS_NODE_DPS_X509_REGISTRATION_18_001: [ `register` shall call `getCertificate` on the security object to acquire the X509 certificate. ] */
            this._securityClient.getCertificate((err, cert) => {
                if (err) {
                    /* Codes_SRS_NODE_DPS_X509_REGISTRATION_18_006: [ If `getCertificate`fails, `register` shall call `_callback` with the error ] */
                    debugErrors('security client returned error on cert acquisition: ' + err);
                    _callback(err);
                }
                else {
                    let registrationId = this._securityClient.getRegistrationId();
                    let request = {
                        registrationId: registrationId,
                        provisioningHost: this._provisioningHost,
                        idScope: this._idScope
                    };
                    /* Codes_SRS_NODE_DPS_X509_REGISTRATION_06_001: [ If `setProvisioningPayload` is invoked prior to invoking `register` than the `payload` property of the `RegistrationRequest` shall be set to the argument provided to the `setProvisioningPayload`.] */
                    if (this._provisioningPayload) {
                        request.payload = this._provisioningPayload;
                    }
                    /* Codes_SRS_NODE_DPS_X509_REGISTRATION_18_004: [ `register` shall pass the certificate into the `setAuthentication` method on the transport ] */
                    this._transport.setAuthentication(cert);
                    /* Codes_SRS_NODE_DPS_X509_REGISTRATION_18_002: [ `register` shall call `registerX509` on the transport object and call it's callback with the result of the transport operation. ] */
                    this._pollingStateMachine.register(request, (err, result) => {
                        this._pollingStateMachine.disconnect((disconnectErr) => {
                            if (disconnectErr) {
                                debugErrors('error disconnecting.  Ignoring.  ' + disconnectErr);
                            }
                            if (err) {
                                /* Codes_SRS_NODE_DPS_X509_REGISTRATION_18_005: [ If `register` on the pollingStateMachine fails, `register` shall call `_callback` with the error ] */
                                _callback(err);
                            }
                            else {
                                _callback(null, result.registrationState);
                            }
                        });
                    });
                }
            });
        }, callback);
    }
    cancel(callback) {
        return (0, azure_iot_common_1.errorCallbackToPromise)((_callback) => {
            this._transport.cancel(_callback);
        }, callback);
    }
}
exports.X509Registration = X509Registration;
//# sourceMappingURL=x509_registration.js.map